import json
import unittest

import numpy as np
import pandas as pd

from welleng.survey import Survey, SurveyHeader
from welleng.utils import get_sigmas

"""
Test that the ISCWSA MWD Rev5 error model is working within a defined
tolerance (the default has been set to 0.001%), testing against the
MWD Rev 5 error model example provided by ISCWSA.
"""

# Set test tolerance as percentage
TOLERANCE = 0.001
# input_files = {
#     "iscwsa_mwd_rev4": "tests/test_data/error_mwdrev4_1_iscwsa_data.json",
#     "iscwsa_mwd_rev5": "tests/test_data/error_mwdrev5_1_iscwsa_data.json"
# }
input_files = {
    "ISCWSA MWD Rev4": "tests/test_data/error_mwdrev4_1_iscwsa_data.json",
    "ISCWSA MWD Rev5": "tests/test_data/error_mwdrev5_1_iscwsa_data.json"
}


def initiate(error_model, filename):
    # Read validation data from file:
    wd = json.load(open(filename))
    err = get_err(error_model, wd)
    df = pd.DataFrame(wd['vd'])

    return (df, err)


def get_md_index(error_data, md):
    i = np.where(error_data.survey.md == md)[0][0]
    return i


def get_err(error_model, wd):
    sh = SurveyHeader()

    for k, v in wd['header'].items():
        setattr(sh, k, v)

    survey = Survey(
        md=wd['survey']['md'],
        inc=wd['survey']['inc'],
        azi=wd['survey']['azi'],
        header=sh,
        error_model=error_model
    )

    err = survey.err

    return err


# initiate lists
class IscwsaMwdErroroTest(unittest.TestCase):
    def test_iscwsa_error_models(self, input_files=input_files):
        for error_model, filename in input_files.items():
            df, err = initiate(error_model, filename)

            nn_c, ee_c, vv_c, ne_c, nv_c, ev_c = [], [], [], [], [], []
            data = [
                nn_c, ee_c, vv_c, ne_c, nv_c, ev_c
            ]

            # generate error data
            for index, row in df.iterrows():
                i = get_md_index(err, row['md'])
                s = row['source']
                if s in ["Totals", "TOTAL"]:
                    source_cov = err.errors.cov_NEVs.T[i]
                else:
                    source_cov = err.errors.errors[s].cov_NEV.T[i]
                v = get_sigmas(source_cov, long=True)
                for j, d in enumerate(v):
                    data[j].append(d[0])

            # convert to dictionary
            ed = {}
            headers = [
                'nn_c', 'ee_c', 'vv_c', 'ne_c', 'nv_c', 'ev_c'
            ]
            for i, h in enumerate(headers):
                ed[h] = data[i]

            df_c = pd.DataFrame(ed)

            df_r = df.join(df_c)

            headers = [
                'nn_d', 'ee_d', 'vv_d', 'ne_d', 'nv_d', 'ev_d'
            ]
            df_d = pd.DataFrame(
                np.around(
                    np.array(df_c) - np.array(df.iloc[:, 2:]),
                    decimals=4
                ),
                columns=headers
            )

            df_r = df_r.join(df_d)

            with np.errstate(divide='ignore', invalid='ignore'):
                error = np.nan_to_num(np.absolute(
                    np.array(df_d) / np.array(df.iloc[:, 2:])
                    ) * 100
                )

            assert np.all(error < TOLERANCE), (
                f"failing error {d}"
            )

            # if you wanted to view the results, this would save then to an Excel
            # file.
            df_r.to_excel(
                "tests/test_data/error_mwdrev5_iscwsa_validation_results.xlsx"
            )


# make above test runnanble separately
if __name__ == '__main__':
    unittest.main()
    # test_iscwsa_error_models(input_files)



import unittest

import welleng as we

survey = we.survey.Survey(
    md=[0, 500, 1000, 2000, 2500, 3500],
    inc=[0, 0, 30, 90, 100, 80],
    azi=[45, 45, 45, 90, 90, 180],
    radius=10,
)


class SurveyInterpolateTest(unittest.TestCase):
    def test_survey_interpolate_survey(self, step=30):
        global survey
        survey_interp = we.survey.interpolate_survey(survey, step=step)
        assert isinstance(survey_interp, we.survey.Survey)

        survey_interp = survey.interpolate_survey(step=step)
        assert isinstance(survey_interp, we.survey.Survey)

    def test_survey_interpolate_survey_tvd(self, step=10):
        global survey
        survey_interp = survey.interpolate_survey(step=30)
        survey_interp_tvd = we.survey.interpolate_survey_tvd(
            survey_interp, step=step
        )
        assert isinstance(survey_interp_tvd, we.survey.Survey)

        survey_interp_tvd = survey_interp.interpolate_survey_tvd(step=step)
        assert isinstance(survey_interp_tvd, we.survey.Survey)

    def test_interpolate_md(self, md=800):
        global survey
        node = survey.interpolate_md(md=md)
        assert isinstance(node, we.node.Node)

    def test_interpolate_tvd(self, tvd=800):
        global survey
        node = survey.interpolate_tvd(tvd=tvd)
        assert isinstance(node, we.node.Node)


# def one_function_to_run_them_all():
#     """
#     Function to gather the test functions so that they can be tested by
#     running this module.

#     https://stackoverflow.com/questions/18907712/python-get-list-of-all-
#     functions-in-current-module-inspecting-current-module
#     """
#     test_functions = [
#         obj for name, obj in inspect.getmembers(sys.modules[__name__])
#         if (inspect.isfunction(obj)
#             and name.startswith('test')
#             and name != 'all')
#     ]

#     for f in test_functions:
#         f()

#         pass


if __name__ == '__main__':
    unittest.main()
    # one_function_to_run_them_all()




"""
test_minimal.py
--------------
Test things that should work with a *minimal* welleng install.
"""
import unittest
import welleng as we


class MinimalTest(unittest.TestCase):

    def test_survey(self):
        survey = we.survey.interpolate_survey(
            survey=we.survey.Survey(
                md=[0, 500, 1000, 2000, 3000],
                inc=[0, 0, 30, 90, 90],
                azi=[90, 90, 90, 135, 180],
                error_model='ISCWSA MWD Rev5'
            ),
            step=30.
        )
        return survey


if __name__ == '__main__':
    unittest.main()



import unittest

import numpy as np
from numpy.typing import NDArray

from welleng.units import ureg
from welleng.utils import (
    annular_volume, decimal2dms, dms2decimal, pprint_dms, dms_from_string
)

LAT, LON = (52, 4, 43.1868, 'N'), (4, 17, 19.6368, 'E')


def _generate_random_dms(n: int, ndigits: int = None) -> NDArray:
    """Generates a bunch of lat, lon coordinates.
    """
    assert n % 2 == 0, "n must be an even int"
    deg = np.random.randint(0, 180, n)
    min = np.random.randint(0, 60, n)
    sec = np.random.uniform(0, 60, n)

    if ndigits is not None:
        sec = np.around(sec, ndigits)

    data = {
        0: 'N', 1: 'S', 2: 'E', 3: 'W'
    }

    direction = np.array([
        data.get(d)
        for d in np.ravel(np.stack((
            np.random.randint(0, 2, int(n/2)),
            np.random.randint(2, 4, int(n/2))
        ), axis=1
        ))
    ])

    return np.stack(
        (deg, min, sec, direction),
        axis=-1,
        dtype=object
    ).reshape((-1, 2, 4))


def are_tuples_identical(tuple1, tuple2):
    return all(x == y for x, y in zip(tuple1, tuple2))


class UtilsTest(unittest.TestCase):
    def test_annular_volume(self):
        av = annular_volume(
            od=ureg('12.25 inch').to('meter'),
            id=ureg(f'{9+5/8} inch').to('meter'),
            length=ureg('1000 meter')
        )

        assert av.m == 3.491531223156194
        assert str(av.u) == 'meter ** 3'

        pass

    def test_decimal2dms(self):
        degrees, minutes, seconds, direction = decimal2dms(
            (LAT[0] + LAT[1] / 60 + LAT[2] / 3600, LAT[3])
        )
        assert (degrees, minutes, round(seconds, 4), direction) == LAT

        dms = decimal2dms(
            (LAT[0] + LAT[1] / 60 + LAT[2] / 3600), ndigits=4
        )
        assert np.all(np.equal(
            dms,
            np.array((np.array(LAT[:3])))
        ))

        dms = decimal2dms(np.array([
            (LAT[0] + LAT[1] / 60 + LAT[2] / 3600),
            (LON[0] + LON[1] / 60 + LON[2] / 3600)
        ]).reshape((-1, 1)), ndigits=4)
        assert np.all(np.equal(
            dms,
            np.array((np.array(LAT[:3]), np.array(LON[:3])))
        ))

        dms = decimal2dms(np.array([
            (LAT[0] + LAT[1] / 60 + LAT[2] / 3600),
            (LON[0] + LON[1] / 60 + LON[2] / 3600)
        ]), ndigits=4)
        assert np.all(np.equal(
            dms,
            np.array((np.array(LAT[:3]), np.array(LON[:3])))
        ))

        dms = decimal2dms(np.array([
            (LAT[0] + LAT[1] / 60 + LAT[2] / 3600, LAT[3]),
            (LON[0] + LON[1] / 60 + LON[2] / 3600, LON[3])
        ]), ndigits=4)
        assert np.all(np.equal(
            dms,
            np.array((np.array(LAT, dtype=object), np.array(LON, dtype=object)))
        ))

    def test_dms2decimal(self):
        decimal = dms2decimal((-LAT[0], LAT[1], LAT[2], LAT[3]))  # check it handles westerly
        assert np.all(np.equal(
            decimal,
            np.array([
                -(LAT[0] + LAT[1] / 60 + LAT[2] / 3600),
                LAT[3]
            ], dtype=object)
        ))

        decimal = dms2decimal((LAT[:3]))
        assert decimal == LAT[0] + LAT[1] / 60 + LAT[2] / 3600

        decimals = dms2decimal((LAT[:3], LON[:3]))
        assert np.all(np.equal(
            decimals,
            np.array((dms2decimal(LAT[:3]), dms2decimal(LON[:3])))
        ))

        decimals = dms2decimal((LAT, LON))
        assert np.all(np.equal(
            decimals,
            np.array((dms2decimal(LAT), dms2decimal(LON))).reshape(decimals.shape)
        ))

        decimals = dms2decimal(((LAT, LON), (LON, LAT)))
        assert np.all(np.equal(
            decimals,
            np.array((
                (dms2decimal(LAT), dms2decimal(LON)),
                (dms2decimal(LON), dms2decimal(LAT))
            ))
        ))

    def test_dms2decimal2dms(self):
        _dms = _generate_random_dms(int(1e3), 8)
        decimal = dms2decimal(_dms)
        dms = decimal2dms(decimal, 8)

        assert np.all(np.equal(_dms, dms))

    def test_pprint_dms(self):
        result = pprint_dms(LAT, return_data=True)
        data = dms_from_string(result)

        assert are_tuples_identical(LAT, data)

        result = pprint_dms(LAT, return_data=True, symbols=False)
        data = dms_from_string(result)

        assert are_tuples_identical(LAT, data)

        result = pprint_dms(LAT[:3], return_data=True)
        data = dms_from_string(result)

        assert are_tuples_identical(LAT[:3], data)

        result = pprint_dms(LAT[:3], return_data=True, symbols=False)
        data = dms_from_string(result)

        assert are_tuples_identical(LAT[:3], data)

        pass


# def one_function_to_run_them_all():
#     """
#     Function to gather the test functions so that they can be tested by
#     running this module.

#     https://stackoverflow.com/questions/18907712/python-get-list-of-all-
#     functions-in-current-module-inspecting-current-module
#     """
#     test_functions = [
#         obj for name, obj in inspect.getmembers(sys.modules[__name__])
#         if (inspect.isfunction(obj)
#             and name.startswith('test')
#             and name != 'all')
#     ]

#     for f in test_functions:
#         f()

#         pass


if __name__ == '__main__':
    unittest.main()
    # one_function_to_run_them_all()



import inspect
import sys
import unittest

import numpy as np

from welleng.connector import Connector
from welleng.survey import Survey, from_connections


class ConnectorTest(unittest.TestCase):
    def test_md_hold(self):
        # test hold with only md provided
        c = Connector(
            vec1=[0, 0, 1],
            md2=500,
        )
        assert (
            c.inc_target == c.inc1
            and c.azi_target == c.azi1
            and c.pos_target[2] == c.md_target
        ), "Failed c1"
        assert c.method == 'hold', "Unexpected method"

        assert isinstance(from_connections(c), Survey)

    def test_md_and_vec(self):
        # test with md2 and vec2 provided (minimum curvature)
        c = Connector(
            vec1=[0, 0, 1],
            md2=1000,
            vec2=[0, 1, 0]
        )
        assert c.method == 'min_curve'

    def test_pos(self):
        # test with pos2 provided (minimum distance)
        c = Connector(
            vec1=[0, 0, 1],
            pos2=[100, 100, 1000],
        )
        assert c.md_target > c.pos1[2], "Failed c3"

    def test_pos_and_dls(self):
        # test with pos2 needing more aggressive dls (minimum curvature)
        c = Connector(
            vec1=[0, 0, 1],
            pos2=[200, 400, 200]
        )
        assert c.method == 'min_curve_to_target'

    def test_pos_and_vec(self):
        # test with pos2 and vec2 provided
        vec1 = [-1, -1, 1]
        vec2 = [1, -1, 0]
        c = Connector(
            pos1=[0., 0., 0],
            vec1=vec1 / np.linalg.norm(vec1),
            pos2=[0., 1000., 500.],
            vec2=vec2 / np.linalg.norm(vec2),
        )
        assert c.method == 'curve_hold_curve'

        # test if interpolator and survey functions are working
        assert isinstance(from_connections(c, step=30), Survey)

    def test_pos_inc_azi(self):
        # test with pos2, inc1 and azi1 provided
        c = Connector(
            pos1=[0., 0., 0],
            inc1=0.,
            azi1=90,
            pos2=[1000., 1000., 1000.],
            vec2=[0., 0., 1.],
        )
        assert c.method == 'curve_hold_curve'

    def test_dls2(self):
        # test with different dls for second curve section
        c = Connector(
            pos1=[0., 0., 0],
            vec1=[0., 0., 1.],
            pos2=[0., 100., 1000.],
            vec2=[0., 0., 1.],
            dls_design2=5
        )
        assert c.radius_design2 < c.radius_design

    def test_radius_critical(self):
        # test with dls_critical requirement (actual dls < dls_design)
        c = Connector(
            pos1=[0., 0., 0],
            vec1=[0., 0., 1.],
            pos2=[0., 100., 100.],
            vec2=[0., 0., 1.],
        )
        assert c.radius_critical < c.radius_design

    def test_min_curve(self):
        # test min_curve (inc2 provided)
        c = Connector(
            pos1=[0., 0., 0],
            vec1=[0., 0., 1.],
            inc2=30,
        )
        assert c.method == 'min_curve'

    def test_radius_critical_with_min_curve(self):
        # test min_curve with md less than required radius
        c = Connector(
            pos1=[0., 0., 0],
            inc1=0,
            azi1=0,
            md2=500,
            inc2=90,
            azi2=0,
        )
        assert c.radius_critical < c.radius_design
# def one_function_to_run_them_all():
#     """
#     Function to gather the test functions so that they can be tested by
#     running this module.

#     https://stackoverflow.com/questions/18907712/python-get-list-of-all-
#     functions-in-current-module-inspecting-current-module
#     """
#     test_functions = [
#         obj for name, obj in inspect.getmembers(sys.modules[__name__])
#         if (inspect.isfunction(obj)
#             and name.startswith('test')
#             and name != 'all')
#     ]

#     [f() for f in test_functions]


if __name__ == '__main__':
    unittest.main()
    # one_function_to_run_them_all()



import unittest

from welleng.fluid import Fluid


class FluidTest(unittest.TestCase):
    def test_fluid_density_profile(self):
        """
        Test that the get_density_profile method in the Fluid class returns the
        same value as the example in the SPE 11118 paper.
        """
        fluid = Fluid(
            fluid_density=10.,  # ppg
            reference_temp=120.,  # Fahrenheit,
            weighting_material='SPE_11118',
            base_fluid_water_ratio=0.103,
        )

        # Override properties to align with ones provided in the example.
        fluid.volume_water_reference_relative = 0.09
        fluid.volume_oil_reference_relative = 0.78
        fluid.volume_weighting_material_relative = 0.11

        density_profile = fluid.get_density_profile(
            depth=10_000.,
            temperature=250.
        )

        assert round(density_profile[-1], 2) == 9.85


if __name__ == '__main__':
    unittest.main()
    # test_fluid_density_profile()



try:
    import trimesh
    TRIMESH = True
except ImportError:
    TRIMESH = False
try:
    import vedo
    from vedo import Lines, Mesh
    from vedo import Plotter as VedoPlotter
    from vedo import Point, Text2D, mag, plotter_instance
    from vedo.addons import Icon, compute_visible_bounds
    from vedo.utils import buildPolyData
    VEDO = True
except ImportError:
    VEDO = False
import numpy as np
from scipy.spatial import KDTree

try:
    import plotly.graph_objects as go
    from plotly.subplots import make_subplots
    PLOTLY = True
except ImportError:
    PLOTLY = False

from vtk import vtkAxesActor, vtkCubeAxesActor, vtkNamedColors

from . import mesh
from .version import __version__ as VERSION

# VEDO = False


class Plotter(VedoPlotter):
    def __init__(self, *args, **kwargs):
        """
        Notes
        -----
        On account of Z or TVD pointing down in the drilling world, the
        coordinate system is right handed. In order to map coordinates in the
        NEV (or North, East, Vertical) reference correctly, North coordinates
        are plotted on the X axis and East on the Y axis. Be mindful of this
        adding objects to a scene.
        """
        super().__init__(*args, **kwargs)

        self.wells = []

        pass

    def add(self, obj, *args, **kwargs) -> None:
        """Modified method to support direct plotting of
        ``welleng.mesh.WellMesh`` instances and for processing the callback
        to print well data when the pointer is hovered of a well trajectory.

        If the ``obj`` is a ``welleng.mesh.WellMesh`` instance, then the args
        and kwargs will be passed to the `vedo.Mesh` instance to facilate e.g.
        color options etc.

        Notes
        -----
        ``welleng.mesh.WellMesh`` creates ``trimesh.Mesh`` instances, a legacy
        of using the ``trimesh`` library for detecting mesh collisions when
        developing automated well trajectory planning. Therefore, to visualize
        the meshes with ``vedo`` and ``vtk``, the meshes need to be converted.

        Meshes in ``welleng`` typically reference an 'NEV' coordinate system,
        which is [North, East, Vertical]. To map correctly to ``vtk``, North
        needs to be mapped to X and East to Y on account of Z pointing down.
        """
        if isinstance(obj, mesh.WellMesh):
            poly = buildPolyData(obj.mesh.vertices, obj.mesh.faces)
            vedo_mesh = Mesh(poly, *args, *kwargs)
            setattr(obj, 'vedo_mesh', vedo_mesh)
            self.wells.append(obj)
            super().add(obj.vedo_mesh)
        else:
            super().add(obj)

        pass

    def _initiate_axes_actor(self):
        plt = vedo.plotter_instance
        r = plt.renderers.index(plt.renderer)

        axact = vtkAxesActor()
        axact.SetShaftTypeToCylinder()
        axact.SetCylinderRadius(0.03)
        axact.SetXAxisLabelText("N")
        axact.SetYAxisLabelText("E")
        axact.SetZAxisLabelText("V")
        axact.GetXAxisShaftProperty().SetColor(1, 0, 0)
        axact.GetYAxisShaftProperty().SetColor(0, 1, 0)
        axact.GetZAxisShaftProperty().SetColor(0, 0, 1)
        axact.GetXAxisTipProperty().SetColor(1, 0, 0)
        axact.GetYAxisTipProperty().SetColor(0, 1, 0)
        axact.GetZAxisTipProperty().SetColor(0, 0, 1)
        bc = np.array(plt.renderer.GetBackground())
        if np.sum(bc) < 1.5:
            lc = (1, 1, 1)
        else:
            lc = (0, 0, 0)
        axact.GetXAxisCaptionActor2D().GetCaptionTextProperty().BoldOff()
        axact.GetYAxisCaptionActor2D().GetCaptionTextProperty().BoldOff()
        axact.GetZAxisCaptionActor2D().GetCaptionTextProperty().BoldOff()
        axact.GetXAxisCaptionActor2D().GetCaptionTextProperty().ItalicOff()
        axact.GetYAxisCaptionActor2D().GetCaptionTextProperty().ItalicOff()
        axact.GetZAxisCaptionActor2D().GetCaptionTextProperty().ItalicOff()
        axact.GetXAxisCaptionActor2D().GetCaptionTextProperty().ShadowOff()
        axact.GetYAxisCaptionActor2D().GetCaptionTextProperty().ShadowOff()
        axact.GetZAxisCaptionActor2D().GetCaptionTextProperty().ShadowOff()
        axact.GetXAxisCaptionActor2D().GetCaptionTextProperty().SetColor(lc)
        axact.GetYAxisCaptionActor2D().GetCaptionTextProperty().SetColor(lc)
        axact.GetZAxisCaptionActor2D().GetCaptionTextProperty().SetColor(lc)
        axact.PickableOff()
        icn = Icon(axact, size=0.1)
        plt.axes_instances[r] = icn
        icn.SetInteractor(plt.interactor)
        icn.EnabledOn()
        icn.InteractiveOff()
        plt.widgets.append(icn)

    def _pointer_callback(self, event):
        i = event.at
        pt2d = event.picked2d
        objs = self.at(i).objects
        pt3d = self.at(i).compute_world_coordinate(pt2d, objs=objs)
        if mag(pt3d) < 0.01:
            if self.pointer is None:
                return
            # if self.pointer in self.at(i).actors:
            self.at(i).remove(self.pointer)
            self.pointer = None
            self.pointer_text.text('')
            self.render()
            return
        if self.pointer is None:
            self.pointer = Point().color('red').pos(pt3d)
        else:
            self.pointer.pos(pt3d)
        self.at(i).add(self.pointer)

        well_data = self._get_closest_well(pt3d, objs)
        if well_data is None:
            self.pointer_text.text(f'point coordinates: {np.round(pt3d, 3)}')
        else:
            survey = well_data.get('well').s
            idx = well_data.get('idx_survey')
            name = survey.header.name
            md = survey.md[idx]
            inc = survey.inc_deg[idx]
            azi_grid = survey.azi_grid_deg[idx]
            self.pointer_text.text(f'''
                well name: {name}\n
                md: {md:.2f}\t inc: {inc:.2f}\t azi: {azi_grid:.2f}\n
                point coordinates: {np.round(pt3d, 3)}
            ''')
        self.render()

    def _well_vedo_meshes(self):
        return [well.vedo_mesh for well in self.wells]

    def _get_closest_well(self, pos, objs) -> dict:
        wells = [
            well for well in self.wells
            if well.vedo_mesh in objs
        ]
        if not bool(wells):
            return

        results = np.zeros((len(wells), 3))
        for i, well in enumerate(wells):
            tree = KDTree(well.vertices.reshape(-1, 3))
            distance, idx_vertices = tree.query(pos)
            results[i] = np.array([distance, idx_vertices, well.n_verts])

        winner = np.argmin(results[:, 0])
        distance, idx_vertices, n_verts = results[winner]

        return {
            'well': wells[winner],
            'distance': distance,
            'idx_vertices': int(idx_vertices),
            'n_verts': int(n_verts),
            'idx_survey': int(idx_vertices // n_verts)
        }

    def show(self, axes=None, *args, **kwargs):
        # check if there's an axes and if so remove them
        if self.axes is not None:
            self.remove(self.axes)

        self._initiate_axes_actor()

        self.add_callback('mouse move', self._pointer_callback)
        self.pointer_text = Text2D("", pos='bottom-right', s=0.5, c='black')
        self.add(self.pointer_text)
        self.pointer = None

        self = super().show(
            viewup=[0, 0, -1], mode=8,
            axes=CubeAxes() if axes is None else axes,
            title=f'welleng {VERSION}',
            *args, **kwargs
        )

        return self


class CubeAxes(vtkCubeAxesActor):
    def __init__(
        self,
        **kwargs
    ):
        super().__init__()

        # # Determine the bounds from the meshes/actors being plotted rounded
        # # up to the nearest 100 units.
        # bounds = np.array(renderer.ComputeVisiblePropBounds())

        # with np.errstate(divide='ignore', invalid='ignore'):
        #     self.bounds = tuple(np.nan_to_num(
        #         (
        #             np.ceil(np.abs(bounds) / 100) * 100
        #         ) * (bounds / np.abs(bounds))
        #     ))

        plt = vedo.plotter_instance
        r = plt.renderers.index(plt.renderer)
        vbb = compute_visible_bounds()[0]
        self.SetBounds(vbb)
        self.SetCamera(plt.renderer.GetActiveCamera())

        namedColors = vtkNamedColors()
        self.colors = {}

        for n in range(1, 4):
            self.colors[f'axis{n}Color'] = namedColors.GetColor3d(
                kwargs.get(f'axis{n}Color', 'Black')
            )

        # self.SetUseTextActor3D(1)

        # self.SetBounds(self.bounds)
        # self.SetCamera(renderer.GetActiveCamera())

        self.GetTitleTextProperty(0).SetColor(self.colors['axis1Color'])
        self.GetLabelTextProperty(0).SetColor(self.colors['axis1Color'])
        self.GetLabelTextProperty(0).SetOrientation(45.0)

        self.GetTitleTextProperty(1).SetColor(self.colors['axis2Color'])
        self.GetLabelTextProperty(1).SetColor(self.colors['axis2Color'])
        self.GetLabelTextProperty(1).SetOrientation(45.0)

        self.GetTitleTextProperty(2).SetColor(self.colors['axis3Color'])
        self.GetLabelTextProperty(2).SetColor(self.colors['axis3Color'])
        self.GetLabelTextProperty(2).SetOrientation(45.0)

        self.SetGridLineLocation(self.VTK_GRID_LINES_FURTHEST)
        for a in ('X', 'Y', 'Z'):
            getattr(self, f'Get{a}AxesLinesProperty')().SetColor(
                namedColors.GetColor3d('Black')
            )
            getattr(self, f'SetDraw{a}Gridlines')(1)
            getattr(self, f'Get{a}AxesGridlinesProperty')().SetColor(
                namedColors.GetColor3d('Grey')
            )
            getattr(self, f'{a}AxisMinorTickVisibilityOff')()

        # self.DrawXGridlinesOn()
        # self.DrawYGridlinesOn()
        # self.DrawZGridlinesOn()
        # self.SetGridLineLocation(self.VTK_GRID_LINES_FURTHEST)

        # self.XAxisMinorTickVisibilityOff()
        # self.YAxisMinorTickVisibilityOff()
        # self.ZAxisMinorTickVisibilityOff()

        units = kwargs.get('units', None)
        self.SetXTitle('N')
        self.SetXUnits(units)
        self.SetYTitle('E')
        self.SetYUnits(units)
        self.SetZTitle('TVD')
        self.SetZUnits(units)
        # self.SetTickLocation(self.VTK_GRID_LINES_FURTHEST)

        # self.ForceOpaqueOff()
        self.SetFlyModeToClosestTriad()
        # Try and prevent scientific numbering on axes
        self.SetLabelScaling(0, 0, 0, 0)
        self.SetXLabelFormat("%.0f")
        self.SetYLabelFormat("%.0f")
        self.SetZLabelFormat("%.0f")

        plt.axes_instances[r] = self
        plt.renderer.AddActor(self)


def plot(
    data,
    names=None,
    colors=None,
    lines=None,
    targets=None,
    arrows=None,
    text=None,
    boxes=None,
    points=None,
    **kwargs
):
    """
    A deprecated wrapper for the Plotter class, maintained only for
    compatability with older versions.

    Parameters
    ----------
    data: a trimesh.Trimesh object or a list of trimesh.Trimesh
    objects or a trmiesh.scene object
    names: list of strings (default: None)
        A list of names, index aligned to the list of well meshes.
    colors: list of strings (default: None)
        A list of color or colors. If a single color is listed then this is
        applied to all meshes in data, otherwise the list of colors is
        indexed to the list of meshes.
    """
    for k, v in locals().items():
        if k != 'data' and k[0] != '_':
            kwargs.update({k: v})

    plt = Plotter(data, **kwargs)

    plt.show()


def get_lines(clearance):
    """
    Add lines per reference well interval between the closest points on the
    reference well and the offset well and color them according to the
    calculated Separation Factor (SF) between the two wells at these points.

    Parameters
    ----------
        clearance: welleng.clearance object

    Returns
    -------
        lines: vedo.Lines object
            A vedo.Lines object colored by the object's SF values.
    """
    assert VEDO, "ImportError: try pip install welleng[easy]"
    c = clearance.sf
    start_points, end_points = clearance.get_lines()
    lines = Lines(start_points, end_points).cmap('hot_r', c, on='cells')
    lines.addScalarBar(title='SF')

    return lines


def figure(obj, type='scatter3d', **kwargs):
    assert PLOTLY, "ImportError: try pip install plotly"
    func = {
        'scatter3d': _scatter3d,
        'mesh3d': _mesh3d,
        'panel': _panel
    }
    fig = func[type](obj, **kwargs)
    return fig


def _panel(survey, **kwargs):
    fig = make_subplots(
        rows=2, cols=2,
        subplot_titles=(
            "Plan",
            f"Vertical Section: {np.degrees(survey.header.vertical_section_azimuth)} deg",
            "WE Section", "NS Section"
        )
        # shared_xaxes=True, shared_yaxes=True
    )

    fig.add_trace(
        go.Scatter(
            x=survey.e,
            y=survey.n,
            mode='lines',
            name='NE Plan',
            showlegend=False
        ),
        row=1, col=1
    )
    fig.add_trace(
        go.Scatter(
            x=survey.vertical_section,
            y=survey.tvd,
            mode='lines',
            name='Plan',
            showlegend=False
        ),
        row=1, col=2
    )
    fig.add_trace(
        go.Scatter(
            x=survey.n,
            y=survey.tvd,
            mode='lines',
            name='NS Section',
            showlegend=False,
        ),
        row=2, col=2
    )
    fig.add_trace(
        go.Scatter(
            x=survey.e,
            y=survey.tvd,
            mode='lines',
            name='WE Section',
            showlegend=False
        ),
        row=2, col=1
    )

    fig.update_layout(
        xaxis=dict(
            title='West-East'
        ),
        yaxis=dict(
            title='North-South'
        ),
        xaxis2=dict(
            title='Outstep'
        ),
        yaxis2=dict(
            title='TVD',
            autorange="reversed",
            matches='y3'
        ),
        xaxis3=dict(
            title='West-East',
            matches='x'
        ),
        yaxis3=dict(
            title='TVD',
            autorange="reversed"
        ),
        xaxis4=dict(
            title='North-South',
            matches='y'
        ),
        yaxis4=dict(
            title='TVD',
            autorange="reversed",
            matches='y3'
        ),
    )

    return fig


def _scatter3d(survey, **kwargs):
    fig = go.Figure()
    fig.add_trace(
        go.Scatter3d(
            x=survey.e,
            y=survey.n,
            z=survey.tvd,
            name='survey',
            mode='lines',
            hoverinfo='skip'
        )
    )
    if not hasattr(survey, "interpolated"):
        survey.interpolated = [False] * len(survey.md)
    if survey.interpolated is None:
        survey.interpolated = [False] * len(survey.md)
    try:
        n, e, v, md, inc, azi = np.array([
            [n, e, v, md, inc, azi]
            for n, e, v, i, md, inc, azi in zip(
                survey.n, survey.e, survey.tvd, survey.interpolated,
                survey.md, survey.inc_deg, survey.azi_grid_deg
            )
            if bool(i) is True
        ]).T
        if n.size:
            text = [
                f"N: {n:.2f}m<br>E: {e:.2f}m<br>TVD: {v:.2f}m<br>"
                + f"MD: {md:.2f}m<br>INC: {inc:.2f}\xb0<br>AZI: {azi:.2f}\xb0"
                for n, e, v, md, inc, azi in zip(n, e, v, md, inc, azi)
            ]
            fig.add_trace(
                go.Scatter3d(
                    x=e,
                    y=n,
                    z=v,
                    name='interpolated',
                    mode='markers',
                    marker=dict(
                        size=5,
                        color='blue',
                    ),
                    text=text,
                    hoverinfo='text'
                )
            )
    except ValueError:
        pass

    try:
        n, e, v, md, inc, azi = np.array([
            [n, e, v, md, inc, azi]
            for n, e, v, i, md, inc, azi in zip(
                survey.n, survey.e, survey.tvd, survey.interpolated,
                survey.md, survey.inc_deg, survey.azi_grid_deg
            )
            if bool(i) is False
        ]).T
        text = [
            f"N: {n:.2f}m<br>E: {e:.2f}m<br>TVD: {v:.2f}m<br>"
            + f"MD: {md:.2f}m<br>INC: {inc:.2f}\xb0<br>AZI: {azi:.2f}\xb0"
            for n, e, v, md, inc, azi in zip(n, e, v, md, inc, azi)
        ]
        fig.add_trace(
            go.Scatter3d(
                x=e,
                y=n,
                z=v,
                name='survey_point',
                mode='markers',
                marker=dict(
                    size=5,
                    color='red',
                ),
                text=text,
                hoverinfo='text'
            )
        )
    except ValueError:
        pass

    fig = _update_fig(fig, kwargs)

    return fig


def _mesh3d(mesh, **kwargs):
    fig = go.Figure()
    vertices = mesh.vertices.reshape(-1, 3)
    faces = np.array(mesh.faces)
    n, e, v = vertices.T
    i, j, k = faces.T
    fig.add_trace(
        go.Mesh3d(
            x=e, y=n, z=v,
            i=i, j=j, k=k,
        )
    )
    if kwargs.get('edges'):
        tri_points = np.array([
            vertices[i] for i in faces.reshape(-1)
        ])
        n, e, v = tri_points.T
        fig.add_trace(
            go.Scatter3d(
                x=e, y=n, z=v,
                mode='lines',
            )
        )
    fig = _update_fig(fig, kwargs)

    return fig


def _update_fig(fig, kwargs):
    """
    Update the fig axis along with any user defined kwargs.
    """
    fig.update_scenes(
        zaxis_autorange="reversed",
        aspectmode='data',
        xaxis=dict(
            title='East (m)'
        ),
        yaxis=dict(
            title='North (m)',
        ),
        zaxis=dict(
            title="TVD (m)"
        )
    )
    for k, v in kwargs.items():
        if k == "layout":
            fig.update_layout(v)
        elif k == "traces":
            fig.update_traces(v)
        else:
            continue

    return fig



import unittest

import welleng as we
import numpy as np

reference = {
    'x': 588319.02, 'y': 5770571.03, 'northing': 5770571.03,
    'easting': 588319.02, 'latitude': 52.077583926214494,
    'longitude': 4.288694821453205, 'convergence': 1.0166440347220762,
    'scale_factor': 0.9996957469340422, 'magnetic_field_intensity': 49381,
    'declination': 2.213, 'dip': -67.199, 'date': '2023-12-16',
    'srs': 'EPSG:23031',
    'wgs84-utm31': [588225.162, 5770360.512]
}

calculator = we.survey.SurveyParameters(reference.get('srs'))


class SurveyParamsTest(unittest.TestCase):
    def test_known_location(self):
        survey_parameters = calculator.get_factors_from_x_y(
            x=reference.get('x'), y=reference.get('y'),
            date=reference.get('date')
        )
        for k, v in survey_parameters.items():
            try:
                assert round(v, 3) == round(reference.get(k), 3)
            except TypeError:
                assert v == reference.get(k)

        pass

    def test_transform_projection_coordinates(self):
        # Convert survey coordinates from UTM31_ED50 to UTM31_WGS84
        coords = np.array((reference.get('easting'), reference.get('northing')))
        result = calculator.transform_coordinates(coords, 'EPSG:32631')
        assert np.allclose(
            result,
            np.array(reference.get('wgs84-utm31'))
        )

        # Try as a list
        result = calculator.transform_coordinates(
            coords.tolist(), 'EPSG:32631'
        )
        assert np.allclose(
            result,
            np.array(reference.get('wgs84-utm31'))
        )

        # Try as a tuple
        result = calculator.transform_coordinates(
            tuple(coords.tolist()), 'EPSG:32631'
        )
        assert np.allclose(
            result,
            np.array(reference.get('wgs84-utm31'))
        )

        result = calculator.transform_coordinates(
            np.array([coords, coords]),
            'EPSG:32631'
        )
        assert np.allclose(
            result,
            np.full_like(result, reference.get('wgs84-utm31'))
        )

        # Try as a list
        result = calculator.transform_coordinates(
            [coords.tolist(), coords.tolist()],
            'EPSG:32631'
        )
        assert np.allclose(
            result,
            np.full_like(result, reference.get('wgs84-utm31'))
        )

        # Try as a tuple
        result = calculator.transform_coordinates(
            (tuple(coords.tolist()), tuple(coords.tolist())),
            'EPSG:32631'
        )
        assert np.allclose(
            result,
            np.full_like(result, reference.get('wgs84-utm31'))
        )

        pass


# def one_function_to_run_them_all():
#     """
#     Function to gather the test functions so that they can be tested by
#     running this module.

#     https://stackoverflow.com/questions/18907712/python-get-list-of-all-
#     functions-in-current-module-inspecting-current-module
#     """
#     test_functions = [
#         obj for name, obj in inspect.getmembers(sys.modules[__name__])
#         if (inspect.isfunction(obj)
#             and name.startswith('test')
#             and name != 'all')
#     ]

#     for f in test_functions:
#         f()

#         pass


if __name__ == '__main__':
    unittest.main()
    # one_function_to_run_them_all()
